% Batch Monte Carlo for application of digital filter
% 
% [y,uy] = BMC(b,a,uba,x,sigma,delta,runs,...)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,delta,runs,...)
% with p_ba = @(runs) a function handle to draw coefficients [b_0,...,b_Nb,a_1,...,a_Na] 
% OR p_ba a covariance matrix (encoding the mutual uncertainties)
%
% varargin is
% phi, theta: ARMA noise model --> eps(n) = sum phi_k*eps(n-k) + sum theta_k*w(n-k) + w(n)
%                                       with w(n) ~ N(0,sigma^2)
% bLow, aLow - coefficients of additional low pass filter
% p - vector of percentiles for which P then contains the quantiles
% delta - upper bound on time domain systematic deviation
% 
% Possible uses of this function are:
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p,delta)
% 
% copyright Sascha Eichstaedt (PTB) 2012
% version 1.1
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%
%%
function [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,varargin)

b = b(:)';
a = a(:)';

PDFhandle = isa(p_ba,'function_handle');
MCSamples = false;

    if ~PDFhandle        
        if isa(p_ba,'numeric');
            [m,~] = size(p_ba);
          if m==runs
              MCSamples = true;
          else
              MCSamples = false;
          end
          if ~MCSamples
            Uba  = p_ba;
            [m,n] = size(Uba);
            if m==1 || n==1
                Uba = diag(Uba.^2);
            end                
            Uba  = 0.5*(Uba+Uba');
            e = eig(Uba);
            if min(e)<0
                fprintf(2,'Covariance matrix is not positiv semidefinite\n');
                y = []; uy = []; P = [];
                return;
            end
            p_ba = @(runs) mvnrnd([b,a(2:end)],Uba,runs);
          end
        else
           fprintf(2,'Invalid input of filter coefficient uncertainty\n');
           y = []; uy = []; P = [];
           return;           
        end
    end


    nb = numel(b);
    na = numel(a);
    
    phi = 0; theta = 0;
    blow = 1; alow = 1;
    p = [0.025,0.975];
    Delta = 0;
    Y = zeros(runs,numel(x));
    defaults = {phi,theta,blow,alow,p,Delta};

    [phi,theta,blow,alow,p,Delta] = mydeal(defaults,varargin{:});
    
    P = zeros(numel(p),numel(x));


    if MCSamples
        TH = p_ba;
    else
        TH = p_ba(runs);
    end
    
    parfor k=1:runs
        bb = TH(k,1:nb); 
        if na > 1
            aa = [1,TH(k,nb+1:end)];
        else
            aa = 1;
        end        
        e  = noiseprocess(phi,theta,sigma,numel(x));
        xx = x + e;
        delta = rand(size(x))*2*Delta - Delta;
        Y(k,:)= filter(bb,aa,filter(blow,alow,xx)) + delta;
    end
    
    y = mean(Y);
    uy= std(Y);

    if nargout > 2
        P = quantile(Y,p);
    end



end
