% Sequential Monte Carlo for filtering in difference equation form
%
% [y,uy] = SMC_F(b,a,uba,x,sigma,delta,runs,...)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,delta,runs,...)
% with p_ba = @(runs) function_to_draw_[b,a] OR uncertainty matrix Uba
%
%
% varargin is
% phi, theta: ARMA noise model --> eps(n) = sum phi_k*eps(n-k) + sum theta_k*w(n-k) + w(n)
%                                       with w(n) ~ N(0,sigma^2)
% bLow, aLow - coefficients of additional low pass filter
% p - vector of percentiles for which P then contains the quantiles
% delta - upper bound on time domain systematic deviation
% Y0 - start values for Monte Carlo Particles
% 
% Possible calls of this function are:
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,bLow)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p,Delta)
% [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p,Delta,Y0)
% 
% copyright Sascha Eichstaedt (PTB) 2012
% version 1.1
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%

function [y,uy,P] = SMC_F(b,a,p_ba,x,sigma,runs,varargin)

b = b(:)';
a = a(:)';

PDFhandle = isa(p_ba,'function_handle');
MCSamples = false;

    if ~PDFhandle        
        if isa(p_ba,'numeric');
            [m,~] = size(p_ba);
          if m==runs
              MCSamples = true;
          else
              MCSamples = false;
          end
          if ~MCSamples
            Uba  = p_ba;
            [m,n] = size(Uba);
            if m==1 || n==1
                Uba = diag(Uba.^2);
            end                
            Uba  = 0.5*(Uba+Uba');
            e = eig(Uba);
            if min(e)<0
                fprintf(2,'Covariance matrix is not positiv semidefinite\n');
                y = []; uy = []; P = [];
                return;
            end
            p_ba = @(runs) mvnrnd([b,a(2:end)],Uba,runs);
          end
        else
           fprintf(2,'Invalid input of filter coefficient uncertainty\n');
           y = []; uy = []; P = [];
           return;           
        end
    end

nb = numel(b);
na = numel(a);


    phi = 0; theta = 0;
    blow = 1; alow = 1;
    p = [0.025,0.975];
    Delta = 0;
    if na>1
        Ytilde = zeros(runs,na-1);
    else
        Ytilde = zeros(runs,1);
    end

    defaults = {phi,theta,blow,alow,p,Delta,Ytilde};

    [phi,theta,blow,alow,p,Delta,Ytilde] = mydeal(defaults,varargin{:});
    
    blow = blow(:); alow = alow(:);
    
    P = zeros(numel(p),numel(x));


    if MCSamples
        Theta = p_ba;
    else
        Theta = p_ba(runs);
    end

    B = Theta(:,1:nb);
    A = Theta(:,nb+1:end);
    clear Theta;

    y = zeros(1,numel(x));
    uy= zeros(1,numel(x));


if ~isequal(blow,1) 
    X = zeros(runs,numel(blow));        
else
    X = zeros(runs,1);
end

if numel(alow)>1
    Xl= zeros(runs,max(nb,numel(alow)-1)); 
else
    Xl = zeros(runs,nb);
end


if theta~=0
    W = zeros(runs,numel(theta));
    MA= true;
    theta = theta(:);
else
    MA = false;
end

if phi~=0
    E = zeros(runs,numel(phi));
    phi = phi(:);
    AR= true;
else
    AR= false;
end

h = waitbar(0,'Sequential MC propagation by difference equation');
set(h,'Name','SMCP-F progress');

for k=1:numel(x)
    w  = randn(runs,1)*sigma;
    if AR && MA
      E  = [ E*phi + W*theta + w, E(:,1:end-1)];
      W  = [w, W(:,1:end-1)];
    elseif AR
      E  = [E*phi + w, E(:,1:end-1)];
    elseif MA
      E  = W*theta + w;
      W  = [w, W(:,1:end-1)];
    else
      E  = w;
    end
      X  = [x(k) + E,  X(:,1:end-1)];
    if numel(alow) > 1
      Xl = [ X*blow - Xl(:,1:numel(alow)-1)*alow(2:end), Xl(:,1:end-1)];
    else              
      Xl = [ X*blow, Xl(:,1:end-1) ];
    end
      
    if numel(a)>1
        Ytilde  = [sum(B.*Xl(:,1:nb),2) - sum(A.*Ytilde,2) , Ytilde(:,1:end-1)];
    else
        Ytilde  = sum(B.*Xl(:,1:nb),2) + rand(runs,1)*2*Delta - Delta;
    end
    Y = Ytilde(:,1) + rand(runs,1)*2*Delta - Delta;
    
    y(k) = mean(Y(:,1));
    uy(k)= std(Y(:,1));
        
    if nargout > 2
        P(:,k) = quantile(Y(:,1),[.025,.975]);
    end
    
    waitbar(k/numel(x),h);
    
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

close(h);

end

