% Monte Carlo for application of digital filter using update formulae for mean, variance
% and (approximated) histogram
%
% [y,uy] = UMC(b,a,uba,x,sigma,delta,runs,...)
% [y,uy,P] = UMC(b,a,p_ba,x,sigma,delta,runs,...)
% with p_ba = @(runs) function_to_draw_[b,a] 
% OR p_ba is a covariance matrix (encoding the mutual uncertainties)
%
%
% varargin is
% phi, theta: ARMA noise model --> eps(n) = sum phi_k*eps(n-k) + sum theta_k*w(n-k) + w(n)
%                                       with w(n) ~ N(0,sigma^2)
% bLow, aLow - coefficients of additional low pass filter
% p - vector of percentiles for which P then contains the quantiles
% delta - upper bound on time domain systematic deviation
% 
% Possible calls of this function are:
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p)
% [y,uy,P] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,bLow,aLow,p,delta)
% 
% copyright Sascha Eichstaedt (PTB) 2012
% copyright on updating formulae parts is by Peter Harris (NPL)
% version 1.1
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%

function [y,uy,P] = UMC(b,a,p_ba,x,sigma,runs,varargin)



nb = numel(b);
b = b(:)';
a = a(:)';

PDFhandle = isa(p_ba,'function_handle');
MCSamples = false;

    if ~PDFhandle        
        if isa(p_ba,'numeric');
            [m,~] = size(p_ba);
          if m==runs
              MCSamples = true;
          else
              MCSamples = false;
          end
          if ~MCSamples
            Uba  = p_ba;
            [m,n] = size(Uba);
            if m==1 || n==1
                Uba = diag(Uba.^2);
            end                
            Uba  = 0.5*(Uba+Uba');
            e = eig(Uba);
            if min(e)<0
                fprintf(2,'Covariance matrix is not positiv semidefinite\n');
                y = []; uy = []; P = [];
                return;
            end
            p_ba = @(runs) mvnrnd([b,a(2:end)],Uba,runs);
          end
        else
           fprintf(2,'Invalid input of filter coefficient uncertainty\n');
           y = []; uy = []; P = [];
           return;           
        end
    end

    phi = 0; theta = 0;
    blow = 1; alow = 1;
    p = [0.025,0.975];
    Delta = 0;
    defaults = {phi,theta,blow,alow,p,Delta};

    [phi,theta,blow,alow,p,Delta] = mydeal(defaults,varargin{:});
    
    P = zeros(numel(p),numel(x));


% ------------ preparations for update formulae ------------

% #######
nbins = 100;
% #######

% phase 1: run small number of MC trials to get starting values
small = 1000;

Y = zeros(small,numel(x));

hw = waitbar(0,'UMC initialisation');

for k=1:small
    th = p_ba(1); 
    bb = th(1:nb); aa = [1,th(nb+1:end)];
    e  = noiseprocess(phi,theta,sigma,numel(x));    
    Y(k,:) = filter(bb,aa, filter(blow,alow,x + e) ) + (rand(1,numel(x))*2*Delta - Delta);    
    waitbar(k/small,hw);
end

close(hw);

% bin edges
ed = zeros(nbins,numel(x));
ymin = min(Y);
ymax = max(Y);

for k=1:numel(x)    
    ed(:,k) = linspace(ymin(k),ymax(k),nbins)';
end



% ----------------- run MC block-wise -----------------------
blocksize = min(runs,1e3);
blocks = ceil(runs/blocksize);
f = zeros(nbins,numel(x)); % histogram
% in Peter Harris' notation: K1 = blocks; K0 = blocksize;



hw = waitbar(0,['UMC running with ',num2str(blocks),' blocks']);

%
for m=1:blocks
    curr_block = min(blocksize, runs - (m-1)*blocksize);    
    Y = zeros(curr_block,numel(x));
    TH= p_ba(blocksize);
  parfor k=1:curr_block
    th = TH(k,:); 
    bb = th(1:nb); aa = [1,th(nb+1:end)];
    e  = noiseprocess(phi,theta,sigma,numel(x));    
    Y(k,:) = filter(bb,aa, filter(blow,alow,x + e) ) + (rand(1,numel(x))*2*Delta - Delta);    
  end
   
  
  if m==1
      y = mean(Y);
      uy= std(Y);
    if nargout > 2
      for k=1:numel(x)
          f(:,k) = histc(Y(:,k),ed(:,k));
      end
      ymin = min(Y);
      ymax = max(Y);
    end
    
  else
     K = (m-1)*blocksize; K0 = curr_block;
     % diff to current calculated mean
      d  = sum( Y - repmat(y,K0,1) )/(K+K0);
     % new mean
      y  = y + d;
     % new variance
      s2 = ((K-1)*uy.^2 + K*d.^2 + ...
          sum( (Y - repmat(y,K0,1)).^2))/(K+K0-1);
      uy = sqrt(s2);
   if nargout > 2
     % update histogram values
     for k=1:numel(x)
        f(:,k)  = f(:,k) + histc(Y(:,k),ed(:,k));
     end
     ymin = min([ymin;Y]);
     ymax = max([ymax;Y]);
   end
  end
  waitbar(m/blocks);  
end

close(hw);


if nargout > 2
    % remove the last frequency, which is always zero
    f = f(1:end-1,:);

    hw = waitbar(0,'UMC credible intervals...');
    
    % replace edge limits by ymin and ymax, resp.
    ed(1,:) = min([ymin;ed(2,:)]);
    ed(end,:) = max([ymax;ed(end-1,:)]);
    p025 = zeros(size(y)); p975 = zeros(size(y));
    for k=1:numel(x)
        e = ed(:,k);
        G = [0; cumsum(f(:,k))/sum(f(:,k))];
        
      % quick fix to ensure strictly increasing G
        iz = find(diff(G) == 0);
        if ~isempty(iz)
           for l = 1:length(iz)
                G(iz+1) = G(iz) + 10*eps;
           end
        end
        
        pcov = linspace(G(1),G(end)-0.95,100)';
        ylow = interp1(G,e,pcov);
        yhgh = interp1(G,e,pcov+0.95);
        lcov = yhgh - ylow;
        imin = find(lcov == min(lcov),1,'first');
        p025(k) = ylow(imin); p975(k) = yhgh(imin);
        waitbar(k/numel(x),hw);
    end       
    close(hw);    
    
    P = [p025(:)';p975(:)'];
    
end





end
