% A very simple example for the application of the sequential and update Monte Carlo methods
% for the calculation of dynamic measurement uncertainty.
%
% Setting: Input signal with white noise of known variance
%          Digital filter with known coefficients and assoc. uncertainties/correlations
%
% copyright Sascha Eichstaedt (PTB) 2012
%
% S. Eichst?dt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%


%% Input signal

% sampling interval
    Ts = 1e-7;
% sampling frequency    
    Fs = 1/Ts;
% noise variance (zero covariance assumed)
   sigma = 1e-3;
% signal
    x = [zeros(1,200),ones(1,800),zeros(1,1000)];
    x = x + randn(size(x))*sigma;
    N = numel(x);
    time = (0:N-1)*Ts;
    
%% measurement system (read from data file)
   
 % IIR filter with coefficent vectors b (numerator) and a (denominator)
 % and associated uncertainty/covariance matrix Uba

   [b,a,Uba] = ReadFromFile('example1_filter.txt',5);
   

%% application of filter and propagation of uncertainty

% number of Monte Carlo trials
runs = 5e3;

% Batch Monte Carlo (only for small number of MC runs)
    [y1,uy1,P1] = BMC(b,a,Uba,x,sigma,runs);
% Sequential Monte Carlo using difference equation
    [y2,uy2,P2] = SMC_F(b,a,Uba,x,sigma,runs);
% Sequential Monte Carlo using state-space equation
    [y3,uy3,P3] = SMC_SS(b,a,Uba,x,sigma,runs);
% Update formulae
    [y4,uy4,P4] = UMC(b,a,Uba,x,sigma,runs);

%% plot results
% Note that results are different due to the small number of MC runs. A large number of runs (>1e5)
% with all methods (incl. BMC) possible only on large scale computers (e.g. work stations). 

figure (1);
plot(time,[x;y1;y2;y3;y4]);
legend('filter input','filter output (BMC)',...
       'filter output (SMC-F)','filter output (SMC-SS)',...
       'filter output (UMC)');
xlabel('time / s');
  
  
figure (2);
plot(time,[uy1;uy2;uy3;uy4]);
legend('uncertainty (BMC)','uncertainty (SMC-F)',...
       'uncertainty (SMC-SS)','uncertainty (UMC)');
xlabel('time / s');


figure (3);
plot(time,P1,'r',time,P2,'b',time,P3,'k',time,P4,'m');
title('95% credible intervals BMC (red), SMC-F (blue), SMC-SS (black) and UMC (mag)');
xlabel('time / s');
