% A simple example for the application of the implementations of GUM-S1 for the calculation
% of dynamic measurement uncertainty. Here the covariance matrix is replaced by a user-defined 
% function to draw samples of the distribution associated with the filter coefficients.
%
% Setting: Input signal with white noise of known variance
%          Digital filter with known coefficients and assoc. uniform distribution
%
% copyright Sascha Eichstaedt (PTB) 2012
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%

%% Input signal

% sampling interval
    Ts = 1e-7;
% sampling frequency    
    Fs = 1/Ts;
% noise variance 
   sigma = 1e-3;
% signal
    x = [zeros(1,200),ones(1,800),zeros(1,1000)];
    x = x + randn(size(x))*sigma;
    N = numel(x);
    time = (0:N-1)*Ts;
    
%% measurement system
  % digital low pass filter 
    [b,a] = butter(4,0.1,'low');
  % with uniform distribution assoc. with its coeffitions  
    b_delta = 1e-5; a_delta = 1e-3;
  % function handle to draw coefficients  
    p_ba = @(runs) [ repmat(b-b_delta,runs,1) + rand(runs,numel(b))*2*b_delta, ...
            repmat(a(2:end)-a_delta,runs,1) + rand(runs,numel(a)-1)*2*a_delta];

%% application of filter and propagation of uncertainty

% number of Monte Carlo trials
runs = 5e3;


% Batch Monte Carlo
    [y1,uy1,P1] = BMC(b,a,p_ba,x,sigma,runs);
% Sequential Monte Carlo using difference equation
    [y2,uy2,P2] = SMC_F(b,a,p_ba,x,sigma,runs);
% Sequential Monte Carlo using state-space equation
    [y3,uy3,P3] = SMC_SS(b,a,p_ba,x,sigma,runs);
% Update formulae
    [y4,uy4,P4] = UMC(b,a,p_ba,x,sigma,runs);


    
%% plot results
% Note that results are different due to the small number of MC runs. A large number of runs (>1e5)
% with all methods (incl. BMC) possible only on large scale computers (e.g. work stations). 

figure (1);
plot(time,[x;y1;y2;y3;y4]);
legend('filter input','filter output (BMC)',...
       'filter output (SMC-F)','filter output (SMC-SS)',...
       'filter output (UMC)');
xlabel('time / s');


figure (2);
plot(time,[uy1;uy2;uy3;uy4]);
legend('uncertainty (BMC)','uncertainty (SMC-F)',...
       'uncertainty (SMC-SS)','uncertainty (UMC)');
xlabel('time / s');


figure (3);
plot(time,P1,'r',time,P2,'b',time,P3,'k',time,P4,'m');
title('95% credible intervals BMC (red), SMC-F (blue), SMC-SS (black) and UMC (mag)');
xlabel('time / s');

