% An example for the application of the implementations of GUM-S1 for the calculation
% of dynamic measurement uncertainty. The noise process is now correlated
% in time and the application of the filter causes a systematic error with known upper bound.
%
% Setting: Input signal with ARMA noise process
%          Digital filter with known coefficients and assoc. uniform distribution
%          Known upper bound for systematic error in time domain (delta)
%
% copyright Sascha Eichstaedt (PTB) 2012
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%

%% Input signal

% sampling interval
    Ts = 1e-7;
% sampling frequency    
    Fs = 1/Ts;

% signal
    x = [zeros(1,200),ones(1,800),zeros(1,1000)];
  % noise process parameters
    phi = 0.3; theta = 0.2;
    sigma = 1e-3;
    
    N = numel(x);
    
    x = x + noiseprocess(phi,theta,sigma,N);
    
    time = (0:N-1)*Ts;
    
%% measurement system
  % digital low pass filter 
    [b,a] = butter(4,0.1,'low');
  % with uniformly distribution assoc. with its coeffitions
    b_delta = 1e-5; a_delta = 1e-3;
  % function handle to draw coefficients  
    p_ba = @(runs) [ repmat(b-b_delta,runs,1) + rand(runs,numel(b))*2*b_delta, ...
            repmat(a(2:end)-a_delta,runs,1) + rand(runs,numel(a)-1)*2*a_delta];
    Delta = 0.1;
        
%% application of filter and propagation of uncertainty

runs = 5e3;


% Batch Monte Carlo (only for small number of MC runs)
    [y1,uy1,P1] = BMC(b,a,p_ba,x,sigma,runs,phi,theta,[],[],[],Delta);
% Sequential Monte Carlo using difference equation
    [y2,uy2,P2] = SMC_F(b,a,p_ba,x,sigma,runs,phi,theta,[],[],[],Delta);    
% Sequential Monte Carlo using state-space equation
    [y3,uy3,P3] = SMC_SS(b,a,p_ba,x,sigma,runs,phi,theta,[],[],[],Delta);   
% Update formulae
    [y4,uy4,P4] = UMC(b,a,p_ba,x,sigma,runs,phi,theta,[],[],[],Delta);


    
%% plot results
% Note that results are different due to the small number of MC runs. A large number of runs (>1e5)
% with all methods (incl. BMC) possible only on large scale computers (e.g. work stations). 

figure (1);
plot(time,[x;y1;y2;y3;y4]);
legend('filter input','filter output (BMC)',...
       'filter output (SMC-F)','filter output (SMC-SS)',...
       'filter output (UMC)');

figure (2);
plot(time,[uy1;uy2;uy3;uy4]);
legend('uncertainty (BMC)','uncertainty (SMC-F)',...
       'uncertainty (SMC-SS)','uncertainty (UMC)');

figure (3);
plot(time,P1,'r',time,P2,'b',time,P3,'k',time,P4,'m');
title('95% credible intervals BMC (red), SMC-F (blue), SMC-SS (black) and UMC (mag)');
