% An example for the application of the implementations of GUM-S1 for the calculation
% of dynamic measurement uncertainty.
%
% input signal with ARMA noise process
% digital deconvolution filter with known coefficients and implicitly known PDF
% known upper bound for systematic error in time domain (delta)

% copyright Sascha Eichstaedt (PTB) 2012
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%


%% Input signal

% sampling interval
    Ts = 1e-4;
% sampling frequency    
    Fs = 1/Ts;

% measurand
    y = [zeros(1,200),ones(1,800),zeros(1,1000)];
    N = numel(y);
    time = (0:N-1)*Ts;
    
%% measurement system

% 4th order system as concatenation of two 2nd order systems

% 
    S0_1 = 1;
    del1 = 0.1;
    f0_1 = 30;
%
    S0_2 = S0_1;
    del2 = 0.02;
    f0_2 = 111;    
    
% Uncertainty of parameters
    std_S0     = S0_1*0.003;
    std_delta1 = del1*0.2;
    std_f01    = f0_1*0.02;
    std_delta2 = del2*0.3;
    std_f02    = f0_2*0.02;
% assign multivariate normal distribution
    pars   = [S0_1,del1,f0_1,del2,f0_2];
    spars  = sqrt(10)*[std_S0,std_delta1,std_f01,std_delta2,std_f02];
    Upars  = diag(spars.^2);
    
%% discretisation of transfer function

    bc1 = pars(1)*(2*pi*pars(3))^2;
    ac1 = [1, 2*pars(2)*2*pi*pars(3), (2*pi*pars(3))^2];

    bc2 =  pars(1)*(2*pi*pars(5))^2;
    ac2 = [1, 2*pars(4)*2*pi*pars(5), (2*pi*pars(5))^2];

    [bd1,ad1] = impinvar(bc1,ac1,Fs);
    [bd2,ad2] = impinvar(bc2,ac2,Fs);

    b = conv(bd1,bd2); a = conv(ad1,ad2);
 
    
%% measured signal

  % noise process parameters
    phi = 0.3; theta = 0.2;
    sigma = 1e-3;
  % observed output signal  
    x = filter(b,a,y) + noiseprocess(phi,theta,sigma,N);

    
%% deconvolution filter

  % compensation filter
    bi = example4_fromPhys2Deconv(pars,Fs);
    ai = 1;
    
  % additional low pass  
    fLow = 100;
    nLow = 4;
    [BLp,ALp] = butter(nLow,2*fLow*Ts);
    
  % regularisation error
    Delta = 1e-3;
    
%% application of filter and propagation of uncertainty

runs = 5e3;

[~,T] = mvnrnd(pars,Upars);
p_ba = @(trials) example4_fromPhys2Deconv(mvnrnd(pars,Upars,trials,T),Fs);


% Batch Monte Carlo
    [y1,uy1,P1] = BMC(bi,ai,p_ba,x,sigma,runs,phi,theta,BLp,ALp,[],Delta);
% Sequential Monte Carlo using difference equation
    [y2,uy2,P2] = SMC_F(bi,ai,p_ba,x,sigma,runs,phi,theta,BLp,ALp,[],Delta);
% Sequential Monte Carlo using state-space equation
    [y3,uy3,P3] = SMC_SS(bi,ai,p_ba,x,sigma,runs,phi,theta,BLp,ALp,[],Delta);
% Update formulae
    [y4,uy4,P4] = UMC(bi,ai,p_ba,x,sigma,runs,phi,theta,BLp,ALp,[],Delta);


    
%% plot results
% Note that results are different due to the small number of MC runs. A large number of runs (>1e5)
% with all methods (incl. BMC) possible only on large scale computers (e.g. work stations). 

figure (1);
plot(time,[x;y1;y2;y3;y4]);
legend('filter input','filter output (BMC)',...
       'filter output (SMC-F)','filter output (SMC-SS)',...
       'filter output (UMC)');
xlabel('time / s');

figure (2);
plot(time,[uy1;uy2;uy3;uy4]);
legend('uncertainty (BMC)','uncertainty (SMC-F)',...
       'uncertainty (SMC-SS)','uncertainty (UMC)');
xlabel('time / s');


figure (3);
plot(time,P1,'r',time,P2,'b',time,P3,'k',time,P4,'m');
title('95% credible intervals BMC (red), SMC-F (blue), SMC-SS (black) and UMC (mag)');
xlabel('time / s');

