% calculate realisations of noise process 
%
% E = noiseprocess(phi,theta,sigma,N)
%
% E = ARMA(p,q) process
% model: E[n] = dot(phi,E[n-1:-1:n-p]) + dot(theta,W[n-1:-1:n-q]) + W(n)
% W(n) ~ N(0,sigma^2)
%
% size(E) = [1,N]
%
% varargin{1} = runs ==> size(E) = [runs,N]
% varargin{2} = memeff ==> if true then be memory efficient, else time efficient
%
% Possible uses of this function are
% E = noiseprocess(phi,theta,sigma,N)
% E = noiseprocess(phi,theta,sigma,N,runs) % draw runs realizations
% E = noiseprocess(phi,theta,sigma,N,runs,memory_efficient_flag) % be memory efficient
%
% copyright Sascha Eichstaedt (PTB) 2012
%
%
% S. Eichstdt, A. Link, P. Harris and C. Elster (2012). 
% Efficient implementation of a Monte Carlo method for uncertainty evaluation in dynamic measurements. 
% Metrologia 49, 401-410. 
%
% DISCLAIMER : 
% ============
% This software was developed at Physikalisch-Technische Bundesanstalt 
% (PTB). The software is made available "as is" free of cost. PTB assumes 
% no responsibility whatsoever for its use by other parties, and makes no 
% guarantees, expressed or implied, about its quality, reliability, safety,
% suitability or any other characteristic. In no event will PTB be liable 
% for any direct, indirect or consequential damage arising in connection 
% with the use of this software.
%

function E = noiseprocess(phi,theta,sigma,N,varargin)


if nargin > 4
    runs = varargin{1};
else
    runs = 1;
end

if nargin > 5 && varargin{2}
    memeff = true;
else
    memeff = false;
end

if memeff

  if theta~=0
    W = zeros(runs,numel(theta));
    MA= true;
  else
    MA = false;
  end
  
    E = zeros(runs,N);
    
    E(:,1) = randn(runs,1)*sigma;
    for k=2:N
        w = randn(runs,1)*sigma;
        if MA
            E(:,k) = E(:,k-1:-1:k-min(numel(phi),k-1))*phi(1:min(numel(phi),k-1)) + ...
                        W*theta(:) + w;
            W = [w, W(:,1:end-1)];                
        else
            E(:,k) = E(:,k-1:-1:k-min(numel(phi),k)+1)*phi(1:min(numel(phi),k)) + w;
        end            
    end
    
else
    W = randn(N,runs)*sigma;    
    E = filter([1,theta],[1,phi],W)';
end
